''' Logic gate circuit elements

    See elements.py for information on how elements are defined.
'''

import numpy as _np

_gap = [_np.nan, _np.nan]
_leadlen = .35
_gateh = 1.
_gatel = .65
_notbubble = .12


def andgate(inputs=2, nand=False, inputnots=[], name='AND'):
    ''' Build n-input AND gate element definitions.

        inputs : number of inputs to gate.
        nand   : invert bubble on output
        inputnots : input numbers (starting at 1) with invert bubble
        name   : Define a name for gate. Only used in documentation currently.
    '''

    AND = {
        'name': name,
        'paths': [[[_leadlen, 0], [_leadlen, -_gateh/2], [_gatel+_leadlen, -_gateh/2]],
                  [[_leadlen, 0], [_leadlen, _gateh/2], [_gatel+_leadlen, _gateh/2]]
                 ],
        'shapes': [{'shape': 'arc',
                    'center': [_gatel+_leadlen, 0],
                    'theta1': -90,
                    'theta2': 90,
                    'width': _gateh,
                    'height': _gateh}
                  ],
        'extend': False,
        'anchors': {'out': [_gatel+_gateh/2+_leadlen*2, 0]}
        }

    if nand:
        AND['shapes'].append(
            {'shape': 'circle',
             'center': [_leadlen+_gatel+_gateh/2+_notbubble, 0],
             'radius': _notbubble})

    # Set distance between inputs. A little larger for 2 input gates.
    if inputs == 2:
        dy = _gateh*.5
    elif inputs == 3:
        dy = _gateh*.33
    else:  # inputs > 3:
        dy = _gateh*.4
        backlen = dy * (inputs-1)

    # Add the inputs and define anchor names
    for i in range(inputs):
        y = (i+1 - (inputs/2+.5)) * dy
        AND['anchors']['in%d' % (inputs-i)] = [0, y]

        if (inputs-i) in inputnots:
            AND['shapes'].append(
                {'shape': 'circle',
                 'center': [_leadlen-_notbubble, y],
                 'radius': _notbubble})

            AND['paths'].append([[0, y], [_leadlen-_notbubble*2, y]])
        else:
            AND['paths'].append([[0, y], [_leadlen, y]])

    # Extended back for large number of inputs
    if inputs > 3:
        AND['paths'].append([[_leadlen, backlen/2+dy/2], [_leadlen, -backlen/2-dy/2]])

    # Output lead
    if nand:
        AND['paths'].append([[0, 0], _gap, [_gatel+_gateh/2+_leadlen+_notbubble*2, 0], [_gatel+_gateh/2+_leadlen*2, 0]])
    else:
        AND['paths'].append([[0, 0], _gap, [_gatel+_gateh/2+_leadlen, 0], [_gatel+_gateh/2+_leadlen*2, 0]])

    return AND


def orgate(inputs=2, nor=False, xor=False, inputnots=[], name='OR'):
    ''' Build n-input OR gate element.

        inputs : number of inputs to gate.
        nor    : invert bubble on output
        xor    : exclusive-or
        inputnots : input numbers (starting at 1) with invert bubble
        name   : Define a name for gate. Only used in documentation currently.
    '''
    # Define OR path as a numpy array
    orflat = .5
    xorgap = .15
    x = _np.linspace(0, _gatel+.05)
    y = x**2
    y = y - max(y)
    y = _np.concatenate(([min(y)], y))   # Combine the flat + parabolic parts
    x = _np.concatenate(([0], x+orflat))

    # Back/input side
    y2 = _np.linspace(min(y), -min(y))
    x2 = -y2**2
    back = min(x2)
    x2 = x2 - back

    # Offset for inputs
    x = x + _leadlen
    x2 = x2 + _leadlen

    if xor:
        x = x + xorgap

    tip = max(x)
    orheight = abs(min(y))

    OR = {
        'name': name,
        'paths': [_np.transpose(_np.vstack((x, y))),
                  _np.transpose(_np.vstack((x, -y))),
                  _np.transpose(_np.vstack((x2, y2)))],
        'shapes': [],
        'extend': False,
        'anchors': {'out': [tip+_leadlen, 0]}
         }

    if xor:
        OR['paths'].append(_np.transpose(_np.vstack((x2+xorgap, y2))))

    if nor:
        OR['shapes'].append(
            {'shape': 'circle',
             'center': [tip+_notbubble, 0],
             'radius': _notbubble})

    # Set distance between inputs. A little larger for 2 input gates.
    if inputs == 2:
        dy = _gateh*.5
    elif inputs == 3:
        dy = _gateh*.33
    else:  # inputs > 3:
        dy = _gateh*.4
        backlen = dy * (inputs-1)

    # Add the inputs and define anchor names
    for i in range(inputs):
        y = (i+1 - (inputs/2+.5)) * dy

        xback = _leadlen - y**2 - back
        if inputs > 3 and ((y > orheight) or (y < -orheight)):
            xback = _leadlen

        OR['anchors']['in%d' % (inputs-i)] = [0, y]

        if (inputs-i) in inputnots:
            OR['shapes'].append(
                {'shape': 'circle',
                 'center': [xback-_notbubble, y],
                 'radius': _notbubble})

            OR['paths'].append([[0, y], [xback-_notbubble*2, y]])
        else:
            OR['paths'].append([[0, y], [xback, y]])

    # Extended back for large number of inputs
    if inputs > 3:
        OR['paths'].append([[_leadlen, backlen/2+dy/2], [_leadlen, orheight]])
        OR['paths'].append([[_leadlen, -backlen/2-dy/2], [_leadlen, -orheight]])

    # Output lead
    if nor:
        OR['paths'].append([[0, 0], _gap, [tip+_notbubble*2, 0], [tip+_leadlen, 0]])
    else:
        OR['paths'].append([[0, 0], _gap, [tip, 0], [tip+_leadlen, 0]])

    return OR


# Define common AND/NANDs
AND2 = andgate(2, name='AND2')
AND3 = andgate(3, name='AND3')
AND4 = andgate(4, name='AND4')
NAND2 = andgate(2, nand=True, name='NAND2')
NAND3 = andgate(3, nand=True, name='NAND3')
NAND4 = andgate(4, nand=True, name='NAND4')

# Define common OR gates
OR2 = orgate(2, name='OR2')
OR3 = orgate(3, name='OR3')
OR4 = orgate(4, name='OR4')
NOR2 = orgate(2, nor=True, name='NOR2')
NOR3 = orgate(3, nor=True, name='NOR3')
NOR4 = orgate(4, nor=True, name='NOR4')
XOR2 = orgate(2, xor=True, name='XOR2')
XOR3 = orgate(3, xor=True, name='XOR3')
XOR4 = orgate(4, xor=True, name='XOR4')
XNOR2 = orgate(2, xor=True, nor=True, name='XNOR2')
XNOR3 = orgate(3, xor=True, nor=True, name='XNOR3')
XNOR4 = orgate(4, xor=True, nor=True, name='XNOR4')

# Inverters/Buffers
BUF = {
    'name': 'BUF',
    'paths': [[[0, 0], [_leadlen, 0], [_leadlen, -_gateh/2], [_gatel+_leadlen, 0], [_leadlen, _gateh/2], [_leadlen, 0]],
              [[_gatel+_leadlen, 0], [_gatel+_leadlen*2, 0]]],
    'extend': False,
    'anchors': {'out': [_gatel+_gateh/2+_leadlen*2, 0]}
     }

NOT = {
    'name': 'NOT',
    'paths': [[[0, 0], [_leadlen, 0], [_leadlen, -_gateh/2], [_gatel+_leadlen, 0], [_leadlen, _gateh/2], [_leadlen, 0]],
              [[_gatel+_leadlen+_notbubble*2, 0], [_gatel+_leadlen*2, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [_gatel+_leadlen+_notbubble, 0],
                'radius': _notbubble}],
    'extend': False,
    'anchors': {'out': [_gatel+_gateh/2+_leadlen*2, 0]}
     }

NOTNOT = {
    'name': 'NOTNOT',
    'paths': [[[0, 0], [_leadlen-_notbubble*2, 0], _gap, [_leadlen, 0], [_leadlen, -_gateh/2], [_gatel+_leadlen, 0], [_leadlen, _gateh/2], [_leadlen, 0]],
              [[_gatel+_leadlen+_notbubble*2, 0], [_gatel+_leadlen*2, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [_gatel+_leadlen+_notbubble, 0],
                'radius': _notbubble},
               {'shape': 'circle',
                'center': [_leadlen-_notbubble, 0],
                'radius': _notbubble}],
    'extend': False,
    'anchors': {'out': [_gatel+_gateh/2+_leadlen*2, 0]}
     }



'''
Electrical element symbols for schematic drawing.

Each element is a dictionary with key/values defining how it should be drawn.

Coordinates are all defined in element coordinates, where the element begins
at [0,0] and is drawn from left to right. The drawing engine will then rotate
and translate the element to its final position. A standard resistor is
1 drawing unit long, and with default lead extension will become 3 units long.

Possible dictionary keys:
    name:  A name string for the element. Currently only used for testing.
    paths: A list of each path line in the element. For example, a capacitor
           has two paths, one for each capacitor "plate". On 2-terminal
           elements, the leads will be automatically extended away from the
           first and last points of the first path, and don't need to
           be included in the path.
    base:  Dictionary defining a base element. For example, the variable
           resistor has a base of resistor, then adds an additional path.
    shapes: A list of shape dictionaries.
            'shape' key can be [ 'circle', 'poly', 'arc', 'arrow' ]
            Other keys depend on the shape as follows.
            circle:
                'center': xy center coordinate
                'radius': radius of circle
                'fill'  : [True, False] fill the circle
                'fillcolor' : color for fill
            poly:
                'xy' : List of xy coordinates defining polygon
                'closed': [True, False] Close the polygon
                'fill'  : [True, False] fill the polygon
                'fillcolor' : color for fill
            arc:
                'center' : Center coordinate of arc
                'width', 'height': width and height of arc
                'theta1' : Starting angle (degrees)
                'theta2' : Ending angle (degrees)
                'angle'  : Rotation angle of entire arc
                'arrow'  : ['cw', 'ccw'] Add an arrowhead, clockwise or counterclockwise
            arrow:
                'start'  : start of arrow
                'end'    : end of arrow
                'headwidth', 'headlength': width and length of arrowhead
    theta: Default angle (in degrees) for the element. Overrides the current
           drawing angle.
    anchors: A dictionary defining named positions within the element. For
             example, the NFET element has a 'source', 'gate', and 'drain'
             anchor. Each anchor will become an attribute of the element class
             which can then be used for connecting other elements.

    extend: [True, False] Extend the leads to fill the full element length.
    move_cur: [True, False] Move the drawing cursor location after drawing.
    color: A matplotlib-compatible color for the element. Examples include
           'red', 'blue', '#34ac92'
    drop: Final location to leave drawing cursor.
    lblloc: ['top', 'bot', 'lft', 'rgt'] default location for text label.
            Defaults to 'top'.
    lblofst: Default distance between element and text label.
    labels: List of (label, pos, align) tuples defining text labels to always draw
            in the element. Align is (horiz, vert) tuple of
            (['center', 'left', right'], ['center', 'bottom', 'top'])
    labels: list of label dictionaries. Each label has keys:
            'label' : string label
            'pos'   : xy position
            'align' : (['center', 'left', right'], ['center', 'bottom', 'top']) alignment
            'size'  : font size
'''

import numpy as _np
import re

_gap = [_np.nan, _np.nan]   # To leave a break in the plot

# Resistor is defined as 1 matplotlib plot unit long.
# When default leads are added, the total length will be three units.
_rh = 0.25      # Resistor height
_rw = 1.0 / 6   # Full (inner) length of resistor is 1.0 data unit


# Drawing Functions
def _cycloid(loops=4, ofst=(0, 0), a=.06, b=.19, norm=True, vertical=False, flip=False):
    ''' Generate a prolate cycloid (inductor spiral) that
        will always start and end at y=0.
        loops = Number of loops
        a, b = parameters. b>a for prolate (loopy) cycloid
        norm = Normalize the length to 1 [True/False]
        vertical, flip = Control the orientation of cycloid
    '''
    yint = _np.arccos(a/b)  # y-intercept
    t = _np.linspace(yint, 2*(loops+1)*_np.pi-yint, num=loops*50)
    x = a*t - b*_np.sin(t)
    y = a - b*_np.cos(t)
    x = x - x[0]  # Shift to start at 0,0

    if norm:
        x = x / (x[-1]-x[0])      # Normalize length to 1

    if flip:
        y = -y

    y = y * (max(y)-min(y))/(_rh)  # Normalize to resistor width

    if vertical:
        x, y = y, x

    x = x + ofst[0]
    y = y + ofst[1]

    path = _np.transpose(_np.vstack((x, y)))
    return path


# Basic elements
RES = {
    'name': 'RES',
    'paths': [[[0, 0], [0.5*_rw, _rh], [1.5*_rw, -_rh], [2.5*_rw, _rh],
               [3.5*_rw, -_rh], [4.5*_rw, _rh], [5.5*_rw, -_rh], [6*_rw, 0]]],
    }

RES_VAR = {
    'name': 'RES_VAR',
    'base': RES,
    'shapes': [{'shape': 'arrow',
                'start': [1.5*_rw, -_rh*2],
                'end': [4.5*_rw, _rw*3.5],
                'headwidth': .12,
                'headlength': .2}],
    'paths': [[[1.5*_rw, -_rh*2], [4.5*_rw, _rw*3.5]]]
     }

RBOX = {
    'name': 'RBOX',
    'paths': [[[0, 0], [0, _rh], [_rw*6, _rh], [_rw*6, -_rh],
               [0, -_rh], [0, 0], _gap, [_rw*6, 0]]]
    }

POT = {
    'name': 'POT',
    'base': RES,
    'shapes': [{'shape': 'arrow',
                'start': [_rw*3, _rw*5],
                'end': [_rw*3, _rw*1.5],
                'headwidth': .15,
                'headlength': .25}],
    'lblloc': 'bot',
    'anchors': {'tap': [_rw*3, _rw*1.5]}
     }

_cap_gap = 0.18
CAP = {   # Straight capacitor
    'name': 'CAP',
    'paths': [[[0, 0], _gap, [0, _rh], [0, -_rh], _gap,
               [_cap_gap, _rh], [_cap_gap, -_rh], _gap,
               [_cap_gap, 0]]],
    'anchors': {'center': [_rh, 0]}
    }

CAP_P = {   # Polarized
    'name': 'CAP_P',
    'base': CAP,
    'labels': [{'label': '+',
                'pos': [-_cap_gap*1.2, _cap_gap]}]
    }

CAP2 = {  # Curved capacitor
    'name': 'CAP2',
    'paths': [[[0, 0], _gap, [0, _rh], [0, -_rh], _gap, [_cap_gap, 0]]],
    'shapes': [{'shape': 'arc',
                'center': [(_cap_gap*1.5), 0],
                'theta1': 120,
                'theta2': -120,
                'width': _cap_gap,
                'height': _rh*2.5}],
    'anchors': {'center': [_rh, 0]}
    }

CAP2_P = {   # Polarized
    'name': 'CAP2_P',
    'base': CAP2,
    'labels': [{'label': '+', 'pos': [-_cap_gap*1.2, _cap_gap]}]
    }

_cap_gap = 0.2
XTAL = {   # Crystal
    'name': 'XTAL',
    'paths': [[[0, 0], _gap, [0, _rh], [0, -_rh], _gap,
              [_cap_gap/2, _rh], [_cap_gap/2, -_rh], [_cap_gap*1.5, -_rh],
              [_cap_gap*1.5, _rh], [_cap_gap/2, _rh], _gap,
              [_cap_gap*2, _rh], [_cap_gap*2, -_rh], _gap, [_cap_gap*2, 0]],
             ],
    }

DIODE = {
    'name': 'DIODE',
    'paths': [[[0, 0], _gap, [_rh*1.4, _rh], [_rh*1.4, -_rh], _gap, [_rh*1.4, 0]]],
    'anchors': {'center': [_rh, 0]},
    'shapes': [{'shape': 'poly',
                'xy': _np.array([[0, _rh], [_rh*1.4, 0], [0, -_rh]]),
                'fill': False}]
    }

DIODE_F = {
    'name': 'DIODE_F',
    'base': DIODE,
    'shapes': [{'shape': 'poly',
                'xy': _np.array([[0, _rh], [_rh*1.4, 0], [0, -_rh]]),
                'fill': True}]
    }

_sd = .1
SCHOTTKY = {  # Schottky diode
    'name': 'SCHOTTKY',
    'base': DIODE,
    'paths': [[[_rh*1.4, _rh], [_rh*1.4-_sd, _rh],  [_rh*1.4-_sd, _rh-_sd]],
              [[_rh*1.4, -_rh], [_rh*1.4+_sd, -_rh], [_rh*1.4+_sd, -_rh+_sd]]]
    }

SCHOTTKY_F = {
    'name': 'SCHOTTKY_F',
    'base': DIODE_F,
    'paths': [[[_rh*1.4, _rh], [_rh*1.4-_sd, _rh],  [_rh*1.4-_sd, _rh-_sd]],
              [[_rh*1.4, -_rh], [_rh*1.4+_sd, -_rh], [_rh*1.4+_sd, -_rh+_sd]]]
    }

ZENER = {  # Zener diode
    'name': 'ZENER',
    'base': DIODE,
    'paths': [[[_rh*1.4, _rh], [_rh*1.4+_sd, _rh+_sd]],
              [[_rh*1.4, -_rh], [_rh*1.4-_sd, -_rh-_sd]]]
    }

ZENER_F = {
    'name': 'ZENER_F',
    'base': DIODE_F,
    'paths': [[[_rh*1.4, _rh], [_rh*1.4+_sd, _rh+_sd]],
              [[_rh*1.4, -_rh], [_rh*1.4-_sd, -_rh-_sd]]]
    }

LED = {
    'name': 'LED',
    'base': DIODE,
    'paths': [[[_rh, _rh*1.5], [_rh*2, _rh*3.25]]],  # Duplicate arrow with a path to work around matplotlib autoscale bug.
    'shapes': [{'shape': 'arrow',
                'start': [_rh, _rh*1.5],
                'end': [_rh*2, _rh*3.25],
                'headwidth': .12,
                'headlength': .2},
               {'shape': 'arrow',
                'start': [_rh*.1, _rh*1.5],
                'end': [_rh*1.1, _rh*3.25],
                'headwidth': .12,
                'headlength': .2}]
    }

# LED with squiggle light
x = _np.linspace(-1, 1)
y = -x*(x-.7)*(x+.7)/2 + _rh*2.5
x = _np.linspace(_rh*.75, _rh*1.25)
theta = 20
c = _np.cos(_np.radians(theta))
s = _np.sin(_np.radians(theta))
m = _np.array([[c, s], [-s, c]])
p = _np.transpose(_np.vstack((x, y)))
p = _np.dot(p, m)
p2 = _np.transpose(_np.vstack((x-.2, y)))
p2 = _np.dot(p2, m)
LED2 = {
    'name': 'LED2',
    'base': DIODE,
    'paths': [p, p2],
    'shapes': [{'shape': 'arrow',
                'start': p[1],
                'end': p[0],
                'headwidth': .07,
                'headlength': .08
                },
                {'shape': 'arrow',
                 'start': p2[1],
                 'end': p2[0],
                 'headwidth': .07,
                 'headlength': .08
                 }]
    }


# Photodiode with squiggle light
x = _np.linspace(-1, 1)
y = -x*(x-.7)*(x+.7)/2 + _rh*2.5
x = _np.linspace(_rh*.75, _rh*1.25)
theta = 20
c = _np.cos(_np.radians(theta))
s = _np.sin(_np.radians(theta))
m = _np.array([[c, s], [-s, c]])
p = _np.transpose(_np.vstack((x, y)))
p = _np.dot(p, m)
p2 = _np.transpose(_np.vstack((x-.2, y)))
p2 = _np.dot(p2, m)

PHOTODIODE = {
    'name': 'PHOTODIODE',
    'base': DIODE,
    'paths': [p, p2],
    'shapes': [{'shape': 'arrow',
                'start': p[-2],
                'end': p[-1],
                'headwidth': .07,
                'headlength': .08
                },
                {'shape': 'arrow',
                 'start': p2[-2],
                 'end': p2[-1],
                 'headwidth': .07,
                 'headlength': .08
                 }]
    }


_mr = 0.2
MEMRISTOR = {
    'name': 'MEMRISTOR',
    'paths': [[[0, _mr*1.25], [_mr*5, _mr*1.25], [_mr*5, _mr*-1.25], [0, _mr*-1.25], [0, _mr*1.25]],
              [[0, 0], [_mr, 0], [_mr, -_mr*.75], [_mr*2, -_mr*.75], [_mr*2, _mr*.75], [_mr*3, _mr*.75], [_mr*3, -_mr*.75], [_mr*4, -_mr*.75], [_mr*4, 0], [_mr*5, 0]],
             ],
    'shapes': [{'shape': 'poly',
                  'xy': [[0, _mr*1.25], [0, -_mr*1.25], [_mr/2, -_mr*1.25], [_mr/2, _mr*1.25]],
                  'fill': True
                 }]
    }

_mrv = .25
MEMRISTOR2 = {
    'name': 'MEMRISTOR2',
    'paths': [[[0, 0], [0, _mrv], [_mr, _mrv], [_mr, -_mrv], [_mr*2, -_mrv], [_mr*2, _mrv],
               [_mr*3, _mrv], [_mr*3, -_mrv], [_mr*4, -_mrv], [_mr*4, _mrv],
               [_mr*5, _mrv], [_mr*5, -_mrv], [_mr*6, -_mrv], [_mr*6, 0],
               [_mr*7, 0]],
             ],
    }

JJ = {   # Josphson Junction
    'name': 'JJ',
    'paths': [[[0, 0], _gap, [-_rh, _rh], [_rh, -_rh], _gap, [_rh, _rh], [-_rh, -_rh], _gap, [0, 0]]],
    }

# Connection dots, lines
_dotr = .075
DOT_OPEN = {
    'name': 'DOT_OPEN',
    'paths': [[[0, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [0, 0],
                'radius': _dotr,
                'fill': True,
                'fillcolor': 'white'}],
    'theta': 0,
    'extend': False,
    }

DOT = {
    'name': 'DOT',
    'paths': [[[0, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [0, 0],
                'radius':_dotr,
                'fill': True,
                'fillcolor': 'black'}],
    'theta': 0,
    'extend': False,
    }

ELLIPSIS = {
    'name': 'ELLIPSIS',
    'shapes': [{'shape': 'circle',
                'center': [.5, 0],
                'radius': _dotr/2,
                'fill': True,
                'fillcolor': 'black'},
               {'shape': 'circle',
                'center': [1, 0],
                'radius': _dotr/2,
                'fill': True,
                'fillcolor': 'black'},
               {'shape': 'circle',
                'center': [1.5, 0],
                'radius': _dotr/2,
                'fill': True,
                'fillcolor': 'black'}],
    'extend': False,
    'drop': [2, 0]
    }


LINE = {'name': 'LINE', 'paths': [_np.array([[0, 0]])]}

# Label centered over a gap
GAP_LABEL = {
    'name': 'LABEL',
    'paths': [[[0, 0], _gap, [1, 0]]],
    'lblloc': 'center',
    'lblofst': 0,
    'color': 'white'
    }

GAP = {
    'base': GAP_LABEL
    }

# Label at the specific location
LABEL = {
    'name': 'LABEL',
    'lblloc': 'center',
    'lblofst': 0,
    'move_cur': False
    }


# Grounds
_gnd_gap = 0.12
_gnd_lead = 0.4
GND = {
    'name': 'GND',
    'paths': [[[0, 0], [0, -_gnd_lead], [-_rh, -_gnd_lead], [_rh, -_gnd_lead],
              _gap, [-_rh*.7, -_gnd_gap-_gnd_lead], [_rh*.7, -_gnd_gap-_gnd_lead],
              _gap, [-_rh*.2, -_gnd_gap*2-_gnd_lead],
              [_rh*.2, -_gnd_gap*2-_gnd_lead]]],
    'move_cur': False,
    'extend': False,
    'theta': 0
    }

GND_SIG = {
    'name': 'GND_SIG',
    'paths': [[[0, 0], [0, -_gnd_lead], [-_rh, -_gnd_lead], [0, -_gnd_lead*2],
              [_rh, -_gnd_lead], [0, -_gnd_lead]]],
    'move_cur': False,
    'extend': False,
    'theta': 0
    }

_chgnd_dx = _rh*.75
_chgnd_dy = _rh
GND_CHASSIS = {
    'name': 'GND_CHASSIS',
    'paths': [[[0, 0], [0, -_gnd_lead], [-_chgnd_dx, -_gnd_lead-_chgnd_dy]],
             [[0, -_gnd_lead], [-_chgnd_dx, -_gnd_lead], [-_chgnd_dx*2, -_gnd_lead-_chgnd_dy]],
             [[0, -_gnd_lead], [_chgnd_dx, -_gnd_lead], [0, -_gnd_lead-_chgnd_dy]]],
    'move_cur': False,
    'extend': False,
    'theta': 0
    }
    
_chgnd_dx = _rh*.75
_chgnd_dy = _rh
VSS = {
    'name': 'VSS',
    'paths': [[[0, 0], [0, -_gnd_lead]],
              [[0, -_gnd_lead], [-_chgnd_dx, -_gnd_lead]],
              [[0, -_gnd_lead], [_chgnd_dx, -_gnd_lead]]],
    'move_cur': False,
    'extend': False,
    'theta': 0
    }

_chgnd_dx = _rh*.75
_chgnd_dy = _rh
VDD = {
    'name': 'VDD',
    'paths': [[[0, 0], [0, _gnd_lead]],
              [[0, _gnd_lead], [-_chgnd_dx, _gnd_lead]],
              [[0, _gnd_lead], [_chgnd_dx, _gnd_lead]]],
    'move_cur': False,
    'extend': False,
    'theta': 0
    }


# Opamp
_oa_back = 2.5
_oa_xlen = _oa_back * _np.sqrt(3)/2
_oa_lblx = _oa_xlen/8
_oa_pluslen = .2
OPAMP_NOSIGN = {
    'name': 'OPAMP_NOSIGN',
    'paths': [[[0, 0], [0, _oa_back/2], [_oa_xlen, 0], [0, -_oa_back/2], [0, 0],
              _gap, [_oa_xlen, 0]],  # Triangle
             ],
    'anchors': {'center': [_oa_xlen/2, 0],
                'in1': [0, _oa_back/4],
                'in2': [0, -_oa_back/4],
                'out': [_oa_xlen, 0]},
    'extend': False,
    }

OPAMP = {
    'name': 'OPAMP',
    'paths': [[[0, 0], [0, _oa_back/2], [_oa_xlen, 0], [0, -_oa_back/2], [0, 0],  _gap,  [_oa_xlen, 0]], # Triangle
              [[_oa_lblx-_oa_pluslen/2, _oa_back/4], [_oa_lblx+_oa_pluslen/2, _oa_back/4]],    # '-' sign
              [[_oa_lblx-_oa_pluslen/2, -_oa_back/4], [_oa_lblx+_oa_pluslen/2, -_oa_back/4]],  # '+' sign
              [[_oa_lblx, -_oa_back/4-_oa_pluslen/2], [_oa_lblx, -_oa_back/4+_oa_pluslen/2]]   # '' sign
             ],
    'anchors': {'center': [_oa_xlen/2, 0],
                'in1': [0, _oa_back/4],
                'in2': [0, -_oa_back/4],
                'out': [_oa_xlen, 0]},
    'extend': False,
    }


# FETs
_fetw = _rw*4
_feth = _rw*5
_fetl = _feth/2
_fet_gap = _rw*1.5
_fetr = _rw*.7  # Radius of "not" bubble
NFET = {
    'name': 'NFET',
    'paths': [[[0, 0], [0, -_fetl], [_fetw, -_fetl], [_fetw, -_fetl-_fetw], [0, -_fetl-_fetw], [0, -2*_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl], [_fetw+_fet_gap, -_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl-_fetw/2], [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]]],
    'extend': False,
    'drop': _np.array([0, -2*_fetl-_fetw]),
    'lblloc': 'lft',
    'anchors': {'source': [0, -2*_fetl-_fetw],
                'drain': [0, 0],
                'gate': [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]}
     }

NFET4 = {
    'name': 'NFET4',
    'paths': [[[0, 0], [0, -_fetl], [_fetw, -_fetl], [_fetw, -_fetl-_fetw], [0, -_fetl-_fetw], [0, -2*_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl], [_fetw+_fet_gap, -_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl-_fetw/2], [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]]],
    'shapes': [{'shape': 'arrow',
                'start': [0, -_fetl-_fetw/2],
                'end': [_fetw, -_fetl-_fetw/2],
                'headwidth': 0.2
                }],
    'extend': False,
    'drop': _np.array([0, -2*_fetl-_fetw]),
    'lblloc': 'lft',
    'anchors': {'source': [0, -2*_fetl-_fetw],
                'drain': [0, 0],
                'gate': [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2],
                'bulk': [0, -_fetl-_fetw/2]}
     }

PFET = {
    'name': 'PFET',
    'paths': [[[0, 0], [0, -_fetl], [_fetw, -_fetl], [_fetw, -_fetl-_fetw], [0, -_fetl-_fetw], [0, -2*_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl], [_fetw+_fet_gap, -_fetl-_fetw]],
              [[_fetw+_fet_gap+_fetr*2, -_fetl-_fetw/2], [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]]],
    'shapes': [{'shape': 'circle',
                'center': [_fetw+_fet_gap+_fetr, -_fetl-_fetw/2],
                'radius': _fetr}],
    'extend': False,
    'drop': _np.array([0, -2*_fetl-_fetw]),
    'lblloc': 'lft',
    'anchors': {'source': [0, 0],
                'drain': [0, -2*_fetl-_fetw],
                'gate': [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]}
     }

PFET4 = {
    'name': 'PFET4',
    'paths': [[[0, 0], [0, -_fetl], [_fetw, -_fetl], [_fetw, -_fetl-_fetw], [0, -_fetl-_fetw], [0, -2*_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl], [_fetw+_fet_gap, -_fetl-_fetw]],
              [[_fetw+_fet_gap, -_fetl-_fetw/2], [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2]]],
    'shapes': [{'shape': 'arrow',
                'start': [_fetw, -_fetl-_fetw/2],
                'end': [0, -_fetl-_fetw/2],
                'headwidth': 0.2
                }],
    'extend': False,
    'drop': _np.array([0, -2*_fetl-_fetw]),
    'lblloc': 'lft',
    'anchors': {'source': [0, 0],
                'drain': [0, -2*_fetl-_fetw],
                'gate': [_fetw+_fet_gap+_fetl+_fetr, -_fetl-_fetw/2],
                'bulk': [0, -_fetl-_fetw/2]}
     }


# BJT transistors
_bjt_r = _rw*3.3   # Radius of BJT circle
_bjt_v = _bjt_r*2/3  # x coord of vertical line
_bjt_v_len = _bjt_r*4/3  # height of vertical line
_bjt_a = _bjt_v_len/4    # Intercept of emitter/collector lines
_bjt_emx = _bjt_v + _bjt_r*.7  # x-coord of emitter exiting circle
_bjt_emy = _bjt_v_len*.7    # y-coord of emitter exiting circle

BJT = {
    'name': 'BJT',
    'paths': [[[0, 0], [_bjt_v, 0]],
              [[_bjt_v, _bjt_v_len/2], [_bjt_v, -_bjt_v_len/2]],
              [[_bjt_v, _bjt_a], [_bjt_emx, _bjt_emy], [_bjt_emx, _bjt_emy+_bjt_a]],
              [[_bjt_v, -_bjt_a], [_bjt_emx, -_bjt_emy], [_bjt_emx, -_bjt_emy-_bjt_a]]],
    'extend': False,
    'drop': _np.array([_bjt_emx, _bjt_emy+_bjt_a]),
    'lblloc': 'rgt',
    'anchors': {'base': [0, 0],
                'collector': [_bjt_emx, _bjt_emy+_bjt_a],
                'emitter': [_bjt_emx, -_bjt_emy-_bjt_a]}
    }

BJT_NPN = {
    'name': 'BJT_NPN',
    'base': BJT,
    'shapes': [{'shape': 'arrow',
                'start': [_bjt_v, -_bjt_a],
                'end': [_bjt_emx, -_bjt_emy],
                'headwidth': 0.2
                }]
    }

BJT_PNP = {
    'name': 'BJT_PNP',
    'base': BJT,
    'shapes': [{'shape': 'arrow',
                'end': [_bjt_v, _bjt_a],
                'start': [_bjt_emx, _bjt_emy],
                'headwidth': 0.2
                }],
    'anchors': {'base': [0, 0],
                'emitter': [_bjt_emx, _bjt_emy+_bjt_a],
                'collector': [_bjt_emx, -_bjt_emy-_bjt_a]}
     }

BJT_NPN_C = {
    'name': 'BJT_NPN_C',
    'base': BJT_NPN,
    'shapes': [{'shape': 'circle',
                'center': [_bjt_r, 0],
                'radius': _bjt_r}]
    }

BJT_PNP_C = {
    'name': 'BJT_PNP_C',
    'base': BJT_PNP,
    'shapes': [{'shape': 'circle',
                'center': [_bjt_r, 0],
                'radius': _bjt_r}]
    }

# BJT's with two collectors
_bjt_2c_dy = -.25
BJT_PNP_2C = {
    'name': 'BJT_PNP_2C',
    'base': BJT_PNP,
    'paths': [[[_bjt_v, -_bjt_a-_bjt_2c_dy], [_bjt_emx, -_bjt_emy-_bjt_2c_dy]]],
    'anchors': {'C2': [_bjt_emx, -_bjt_emy-_bjt_2c_dy]}
}

# Inductor without spiraling
_ind_w = .25
_ind_shape_list = []
for _i in range(4):
    _ind_shape_list.append({'shape': 'arc',
                            'center': [(_i*2+1)*_ind_w/2, 0],
                            'theta1': 0,
                            'theta2': 180,
                            'width': _ind_w,
                            'height': _ind_w})
INDUCTOR = {
    'name': 'INDUCTOR',
    'paths': [[[0, 0], _gap, [1, 0]]],
    'shapes': _ind_shape_list}

INDUCTOR2 = {
    'name': 'INDUCTOR2',
    'paths': [_cycloid(loops=4)]
    }

# Independent sources
SOURCE = {
    'name': 'SOURCE',
    'paths': [[[0, 0], [0, 0], _gap, [1, 0], [1, 0]]],
    'theta': 90.,
    'shapes': [{'shape': 'circle',
                'center': [0.5, 0],
                'radius': 0.5}],
     }

_plus_len = .2
SOURCE_V = {
    'name': 'SOURCE_V',
    'base': SOURCE,
    'paths': [[[.25, -_plus_len/2], [.25, _plus_len/2]],     # '-' sign
              [[.75-_plus_len/2, 0], [.75+_plus_len/2, 0]],  # '+' sign
              [[.75, -_plus_len/2], [.75, _plus_len/2]],    # '+' sign
              ]
    }

SOURCE_I = {
    'name': 'SOURCE_I',
    'base': SOURCE,
    'shapes': [{'shape': 'arrow',
                'start': [.75, 0],
                'end': [.25, 0]}]
    }


_sin_y = _np.linspace(.25, .75, num=25) - 0.5
_sin_x = .2 * _np.sin((_sin_y-.25)*_np.pi*2/.5) + 0.5
_sin_path = _np.transpose(_np.vstack((_sin_x, _sin_y)))
SOURCE_SIN = {
    'name': 'SOURCE_SIN',
    'base': SOURCE,
    'paths': [_sin_path]
    }


#Controlled sources
SOURCE_CONT = {
    'name': 'SOURCE_CONT',
    'paths': [[[0, 0], [.5, .5], [1, 0]],
              [[0, 0], [.5, -.5], [1, 0]]],
    'theta': 90.,
     }

SOURCE_CONT_V = {
    'name': 'SOURCE_CONT_V',
    'base': SOURCE_CONT,
    'paths': [[[.25, -_plus_len/2], [.25, _plus_len/2]],     # '-' sign
              [[.75-_plus_len/2, 0], [.75+_plus_len/2, 0]],  # '+' sign
              [[.75, -_plus_len/2], [.75, _plus_len/2]],    # '+' sign
             ]
    }

SOURCE_CONT_I = {
    'name': 'SOURCE_CONT_I',
    'base': SOURCE_CONT,
    'shapes': [{'shape': 'arrow',
                'start': [.75, 0],
                'end': [.25, 0]}]
    }

# Batteries
_batw = _rh*.75
_bat1 = _rh*1.5
_bat2 = _rh*.75
BAT_CELL = {
    'name': 'BAT_CELL',
    'paths': [[[0, _bat1], [0, -_bat1]],
              [[_batw, _bat2], [_batw, -_bat2]],
              [[0, 0], _gap, [_batw, 0]]
             ]
    }

BATTERY = {
    'name': 'BATTERY',
    'paths': [[[0, _bat1], [0,-_bat1]],
              [[_batw, _bat2], [_batw, -_bat2]],
              [[_batw*2, _bat1], [_batw*2, -_bat1]],
              [[_batw*3, _bat2], [_batw*3, -_bat2]],
              [[0, 0], _gap, [_batw*3, 0]]
             ]
    }


# Meters
METER_V = {
    'name': 'METER_V',
    'base': SOURCE,
    'labels': [{'label': 'V', 'pos': [.5, 0]}]
    }

METER_I = {
    'name': 'METER_I',
    'base': SOURCE,
    'labels': [{'label': 'I', 'pos': [.5, 0]}]
    }

METER_OHM = {
    'name': 'METER_OHM',
    'base': SOURCE,
    'labels': [{'label': '$\Omega$', 'pos': [.5, 0]}]
    }


# Arrows for labeling current
ARROWLINE = {
    'name': 'ARROWLINE',
    'paths': [[[0, 0], [1, 0]]],
    'shapes': [{'shape': 'arrow',
                'start': [0, 0],
                'end': [0.55, 0],
                'headwidth': .2,
                'headlength': .3}],
    'lblofst': .2,
    }

ARROW_I = {
    'name': 'ARROW_I',
    'shapes': [{'shape': 'arrow',
                'start': [0, 0],
                'end': [2, 0],
                'headwidth': .2,
                'headlength': .3}],
    'anchors': {'center': [1, 0]},
    'lblofst': .2,
    'move_cur': False
    }

# Switches
_sw_dot_r = .12
SWITCH_SPST = {
    'name': 'SWITCH_SPST',
    'paths': [[[0, 0], _gap, [_sw_dot_r*2, .1], [.8, .45], _gap, [1, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [_sw_dot_r, 0],
                'radius': _sw_dot_r},
               {'shape': 'circle',
                'center': [1-_sw_dot_r, 0],
                'radius': _sw_dot_r}]
    }

SWITCH_SPST_OPEN = {
    'name': 'SWITCH_SPST_OPEN',
    'base': SWITCH_SPST,
    'shapes': [{'shape': 'arc',
                'center': [.4, .1],
                'width': .5,
                'height': .75,
                'theta1': -10,
                'theta2': 70,
                'arrow': 'ccw'}]
    }

SWITCH_SPST_CLOSE = {
    'name': 'SWITCH_SPST_CLOSE',
    'base': SWITCH_SPST,
    'shapes': [{'shape': 'arc',
                'center': [.4, .25],
                'width': .5,
                'height': .75,
                'theta1': -10,
                'theta2': 70,
                'arrow': 'cw'}]
    }

SWITCH_SPDT = {
    'name': 'SWITCH_SPDT',
    'base': SWITCH_SPST,
    'shapes': [{'shape': 'circle',
                'center': [1-_sw_dot_r, .7],
                'radius': _sw_dot_r}],
    'anchors': {'a': [0, 0],
                'b': [1, 0],
                'c': [1, .7]},
    'extend': False
    }

SWITCH_SPDT_OPEN = {
    'name': 'SWITCH_SPDT_OPEN',
    'base': SWITCH_SPST_OPEN,
    'shapes': [{'shape': 'circle',
                'center': [1-_sw_dot_r, .7],
                'radius': _sw_dot_r}],
    'extend': False
    }

SWITCH_SPDT_CLOSE = {
    'name': 'SWITCH_SPDT_CLOSE',
    'base': SWITCH_SPST_CLOSE,
    'shapes': [{'shape': 'circle',
                'center': [1-_sw_dot_r, .7],
                'radius': _sw_dot_r}],
    'extend': False
    }

SWITCH_SPDT2 = {
    'name': 'SWITCH_SPDT2',
    'paths': [[[0, 0], _gap, [_sw_dot_r*2, .1], [.7, .25], _gap, [1, .4]]],
    'shapes': [{'shape': 'circle',
                'center': [_sw_dot_r, 0],
                'radius': _sw_dot_r},
               {'shape': 'circle',
                'center': [1-_sw_dot_r, -.4],
                'radius': _sw_dot_r},
               {'shape': 'circle',
                'center': [1-_sw_dot_r, .4],
                'radius': _sw_dot_r}],
    'anchors': {'a': [0, 0],
                'b': [1, .4],
                'c': [1, -.4]},
    'extend': False
    }

SWITCH_SPDT2_OPEN = {
    'name': 'SWITCH_SPDT2_OPEN',
    'base': SWITCH_SPDT2,
    'shapes': [{'shape': 'arc',
                'center': [.35, 0],
                'width': .5,
                'height': .75,
                'theta1': -10,
                'theta2': 70,
                'arrow': 'ccw'}]
    }

SWITCH_SPDT2_CLOSE = {
    'name': 'SWITCH_SPDT2_CLOSE',
    'base': SWITCH_SPDT2,
    'shapes': [{'shape': 'arc',
                'center': [.3, .1],
                'width': .5,
                'height': .75,
                'theta1': -10,
                'theta2': 70,
                'arrow': 'cw'}]
    }

# Pushbuttons
BUTTON = {
    'name': 'BUTTON',
    'paths': [[[0, 0], _gap, [_sw_dot_r, .3], [1-_sw_dot_r, .3],
               _gap, [.5, .3], [.5, .5], _gap, [1, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [_sw_dot_r, 0],
                'radius': _sw_dot_r},
               {'shape': 'circle',
                'center': [1-_sw_dot_r, 0],
                'radius': _sw_dot_r}]
    }

BUTTON_NC = {
    'name': 'BUTTON_NC',
    'paths': [[[0, 0], _gap, [_sw_dot_r, -_sw_dot_r-.05],
               [1-_sw_dot_r, -_sw_dot_r-.05], _gap, [.5, -_sw_dot_r-.05],
               [.5, _sw_dot_r+.15], _gap, [1, 0]]],
    'shapes': [{'shape': 'circle',
                'center': [_sw_dot_r, 0],
                'radius': _sw_dot_r},
               {'shape': 'circle',
                'center': [1-_sw_dot_r, 0],
                'radius': _sw_dot_r}]
    }


# Speakers, bulbs, etc.
_sph = .5
SPEAKER = {
    'name': 'SPEAKER',
    'paths': [[[0, 0], [_rh, 0]],
              [[0, -_sph], [_rh, -_sph]]],
    'drop': [0, -_sph],
    'extend': False,
    'shapes': [{'shape': 'poly',
                'xy': [[_rh, _sph/2], [_rh, -_sph*1.5], [_rh*2, -_sph*1.5], [_rh*2, _sph/2]]},
               {'shape': 'poly',
                'xy': [[_rh*2, _sph/2], [_rh*3.5, _sph*1.25], [_rh*3.5, -_sph*2.25], [_rh*2, -_sph*1.5]],
                'closed': False}]
    }

_a = .25
_b = .7
_t = _np.linspace(1.4, 3.6*_np.pi, 100)
_x = _a*_t - _b*_np.sin(_t)
_y = _a - _b * _np.cos(_t)
_x = (_x - _x[0])  # Scale to about the right size
_x = _x / _x[-1]
_y = (_y - _y[0]) * .25
_lamp_path = _np.transpose(_np.vstack((_x, _y)))
LAMP = {
    'name': 'LAMP',
    'base': SOURCE,
    'paths': [_lamp_path]
    }


def transformer(t1=4, t2=4, core=True, ltaps=None, rtaps=None, loop=False):
    ''' Transformer element
        t1 = Turns on primary (left) side
        t2 = Turns on secondary (right) side
        core = Draw the core (parallel lines) [default=True]
        ltaps = Dictionary of name:position pairs, position is the turn number from the top to tap
                Each tap defines an anchor point but does not draw anything.
        rtaps = Same as ltaps, on right side
        loop = Use spiral/cycloid (loopy) style [default=False]
    '''
    # Set initial parameters
    ind_w = .4
    lbot = 0
    ltop = t1*ind_w
    rtop = (ltop+lbot)/2 + t2*ind_w/2
    rbot = (ltop+lbot)/2 - t2*ind_w/2

    element = {'name': 'TRANSFORMER',
               'extend': False,
               'paths': []}

    # Adjust for loops or core
    ind_gap = .75
    if loop:
        ind_gap = ind_gap + .4
    if core:
        ind_gap = ind_gap + .25

    ltapx = 0
    rtapx = ind_gap

    # Draw coils
    if loop:
        c1 = _cycloid(loops=t1, ofst=(0, 0), norm=False, vertical=True)
        c2 = _cycloid(loops=t2, ofst=(ind_gap, -rtop+ltop), norm=False, flip=True, vertical=True)
        ltapx = min([i[0] for i in c1])
        rtapx = max([i[0] for i in c2])
        ltop = c1[-1][1]
        rtop = c2[-1][1]
        element['paths'].extend([c1, c2])
    else:
        shapes = []
        for i in range(t1):
            shapes.append({'shape': 'arc',
                           'center': [0, ltop-(i*ind_w+ind_w/2)],
                           'theta1': 270,
                           'theta2': 90,
                           'width': ind_w,
                           'height': ind_w})
        for i in range(t2):
            shapes.append({'shape': 'arc',
                           'center': [ind_gap, rtop-(i*ind_w+ind_w/2)],
                           'theta1': 90,
                           'theta2': 270,
                           'width': ind_w,
                           'height': ind_w})
        element['shapes'] = shapes

    # Add the core
    if core:
        top = max(ltop, rtop)
        bot = min(lbot, rbot)
        center = ind_gap/2
        core_w = ind_gap/10
        element['paths'].extend([[[center-core_w, top], [center-core_w, bot]],
                                 [[center+core_w, top], [center+core_w, bot]]])

    anchors = {'p1': [0, ltop],
               'p2': [0, lbot],
               's1': [ind_gap, rtop],
               's2': [ind_gap, rbot]}

    if ltaps:
        for name, pos in ltaps.items():
            anchors[name] = [ltapx, ltop - pos * ind_w]
    if rtaps:
        for name, pos in rtaps.items():
            anchors[name] = [rtapx, rtop - pos * ind_w]

    element['anchors'] = anchors
    return element


def blackbox(w, h, linputs=None, rinputs=None, tinputs=None, binputs=None,
             mainlabel='', leadlen=0.5, lblsize=16, lblofst=.15, plblofst=.1, plblsize=12,
             hslant=None, vslant=None):
    ''' Generate a black-box element consisting of rectangle with inputs on each side.

        w, h : width and height of rectangle

        linputs, rinputs, tinputs, binputs: dictionary input definition for each side
        of the box. Default to no inputs. Dictionary keys:
            labels: list of string labels for each input. drawn inside the box. default is blank.
            plabels: list of pin label strings. drawn outside the box. Default is blank.
            loc: list of pin locations (0 to 1), along side. Defaults to evenly spaced pins.
            leads: True/False, draw leads coming out of box. Default=True.
            lblofst: label offset override for this side.
            plblofst: pin label offset override for this side.
            lblsize: font size override for labels on this side
            plblsize: font size override for pin labels on this side

        mainlabel : main box label
        leadlen   : length of lead extensions
        lblsize   : default font size of labels
        lblofst   : default label offset
        plblsize  : default pin label size
        plblofst  : default pin label offset
        hslant    : angle (degrees) to slant horizontal sides
        vslant    : angle (degrees) to slant vertical sides

        Anchors will be generated on each pin matching label names if provided. If no
        labels, anchors are named 'inL1', 'inL2', ... 'inR1', 'inR2', 'inT1', 'inB1', etc.
    '''
    if hslant is not None and vslant is not None:
        print('Warning - hslant and vslant both deifined. Weird pins may result.')

    if hslant is None:
        hslant = 0
    if vslant is None:
        vslant = 0

    y1 = 0 - w * _np.tan(_np.deg2rad(hslant))
    y2 = h + w * _np.tan(_np.deg2rad(hslant))
    x1 = 0 - h * _np.tan(_np.deg2rad(vslant))
    x2 = w + h * _np.tan(_np.deg2rad(vslant))
    hleft = h
    hrght = y2-y1
    wbot = w
    wtop = x2-x1

    box = {'paths': [[[0, 0], [w, y1], [x2, y2], [x1, h], [0, 0]]],
           'anchors': {},
           'labels': [],
           'extend': False}

    if mainlabel is not None:
        box['labels'].append({'label': mainlabel, 'pos': [w/2, h*2/3]})

    for side, sidelabel in [('left', linputs), ('right', rinputs), ('top', tinputs), ('bottom', binputs)]:
        if sidelabel is None: continue
        cnt = len(sidelabel.get('labels', []))
        sidelabels = sidelabel.get('labels', [])
        plabels = sidelabel.get('plabels', [])
        leads = sidelabel.get('leads', True)
        _lblofst = sidelabel.get('lblofst', lblofst)
        _plblofst = sidelabel.get('plblofst', plblofst)
        _lblsize = sidelabel.get('lblsize', lblsize)
        _plblsize = sidelabel.get('plblsize', plblsize)

        # Determine pin spacings
        if side == 'left':
            x0 = 0
            if cnt > 1:
                y0 = hleft/(cnt+2)
            else:
                y0 = hleft/2
            lblpos = _np.array([_lblofst, 0])
            lblalign = ('left', 'center')
            plblpos = _np.array([-_plblofst, 0])
            plblalign = ('right', 'bottom')
            leadext = _np.array([-leadlen, 0])
        elif side == 'right':
            x0 = w
            if cnt > 1:
                y0 = hrght/(cnt+2)
                y0 = y0 - (hrght-hleft)
            else:
                y0 = hrght/2
            lblpos = _np.array([-_lblofst, 0])
            lblalign = ('right', 'center')
            plblpos = _np.array([_plblofst, 0])
            plblalign = ('left', 'bottom')
            leadext = _np.array([leadlen, 0])
        elif side == 'top':
            y0 = h
            if cnt > 1:
                x0 = wtop/(cnt+2)
            else:
                x0 = wtop/2
            lblpos = _np.array([0, -_lblofst])
            lblalign = ('center', 'top')
            plblpos = _np.array([_plblofst, 0])
            plblalign = ('left', 'bottom')
            leadext = _np.array([0, leadlen])
        elif side == 'bottom':
            y0 = 0
            if cnt > 1:
                x0 = wbot/(cnt+2)
            else:
                x0 = wbot/2
            lblpos = _np.array([0, _lblofst])
            lblalign = ('center', 'bottom')
            plblpos = _np.array([_plblofst, -_plblofst])
            plblalign = ('left', 'top')
            leadext = _np.array([0, -leadlen])

        if 'loc' not in sidelabel:
            if cnt == 1:
                sidelabel['loc'] = [0.5]
            else:
                dim = {'left': hleft, 'right': hrght, 'top': wtop, 'bottom': wbot}[side]
                sidelabel['loc'] = _np.linspace(1/(cnt+2), 1-1/(cnt+2), num=cnt)
                if 'spacing' in sidelabel:
                    sidelabel['loc'] = (sidelabel['loc'] - 0.5) * (sidelabel['spacing']/dim)/(sidelabel['loc'][1]-sidelabel['loc'][0]) + 0.5

        # Add each input
        for i in range(cnt):
            # Custom-spaced labels
            if side == 'top':
                x = sidelabel['loc'][i] * wtop - (wtop-wbot)
                y = y0
            elif side == 'bottom':
                x = sidelabel['loc'][i] * wbot
                y = y0
            elif side == 'left':
                y = sidelabel['loc'][i] * hleft
                x = x0
            elif side == 'right':
                y = sidelabel['loc'][i] * hrght - (hrght-hleft)
                x = x0
            else:
                assert False

            if x0 == 0:
                x = x - y * _np.tan(_np.deg2rad(vslant))
            else:
                x = x - y * _np.tan(-_np.deg2rad(vslant))

            if y0 == 0:
                y = y - x * _np.tan(_np.deg2rad(hslant))
            else:
                y = y - x * _np.tan(-_np.deg2rad(hslant))

            # Anchor name use label for pin, otherwise use 'inL1', etc.
            if len(sidelabels) == cnt:
                aname = re.sub(r'\W+', '', sidelabels[i])
            else:
                aname = 'in' + side[0].upper() + str(i+1)

            if aname in box['anchors']: aname = aname + '_%d' % i

            # Add lead lines
            if leads:
                box['paths'].append([[x, y], _np.array([x, y])+leadext])
                box['anchors'][aname] = _np.array([x, y])+leadext
            else:
                box['anchors'][aname] = [x, y]

            # Add labels
            if len(sidelabels) == cnt:
                box['labels'].append({'label': sidelabels[i], 'pos': _np.array([x, y])+lblpos, 'align': lblalign, 'size': _lblsize})

            if len(plabels) == cnt:
                box['labels'].append({'label': plabels[i], 'pos': _np.array([x, y])+plblpos, 'align': plblalign, 'size': _plblsize})

    return box


def mux(inputs=None, outputs=None, ctrls=None, topctrls=None,
        demux=False, h=None, w=None,
        pinspacing=1, ctrlspacing=.6, slope=25, **kwargs):
    ''' Generate a multiplexer/demultiplexer element.

        Parameters
        ----------
        inputs: list of strings
            Name of each input, labeled from bottom up
        outputs: list of strings
            Name of each output, labeled from bottom up
        ctrls: list of strings
            Name of control signals (bottom)
        topctrls: list of strings
            Name of control signals on top side
        demux: boolean
            Draw as demultiplexer
        h: float, optional
            Height of multiplexer
        w: float, optional
            Width of multiplexer
        pinspacing: float
            distance between pins on input/output side
        ctrlspacing: float
            distance between pins on control side
        slope: float
            angle (degrees) to slope top and bottom
        **kwargs:
            keyword arguments to pass to blackbox method

        Returns
        -------
        element dictionary
    '''
    linputs = None
    rinputs = None
    binputs = None
    tinputs = None
    wbot = 0
    wtop = 0
    hleft = 0
    hrght = 0
    if not demux:
        slope = -slope

    if inputs is not None:
        hleft = (len(inputs)+1) * max(.4, pinspacing)
        linputs = {'labels': inputs, 'spacing': pinspacing}

    if outputs is not None:
        hrght = (len(outputs)) * max(.4, pinspacing)
        rinputs = {'labels': outputs, 'spacing': pinspacing}

    if ctrls is not None:
        wbot = (len(ctrls)+1) * max(.8, ctrlspacing)
        binputs = {'labels': ctrls, 'spacing': ctrlspacing}

    if topctrls is not None:
        wtop = (len(topctrls)+1) * max(.8, ctrlspacing)
        tinputs = {'labels': topctrls, 'spacing': ctrlspacing}

    h = max(hleft, hrght) if h is None else h
    w = max(wbot, wtop) if w is None else w

    return blackbox(w, h, linputs=linputs, rinputs=rinputs,
                    binputs=binputs, tinputs=tinputs,
                    hslant=slope, **kwargs)



"""
Electrical Schematic Drawing

https://cdelker.bitbucket.io/SchemDraw/
"""

import matplotlib.pyplot as plt
import numpy as np
import matplotlib as mpl
from matplotlib.patches import Arc
import copy

from . import elements

#--------------------------------------------------------------------
# Set up matplotlib parameters
#--------------------------------------------------------------------
mpl.rcParams['figure.subplot.left']   = 0.05
mpl.rcParams['figure.subplot.bottom'] = 0.05
mpl.rcParams['figure.subplot.right']  = 0.95
mpl.rcParams['figure.subplot.top']    = 0.90
mpl.rcParams['font.family'] = 'sans-serif'
mpl.rcParams['mathtext.fontset'] = 'stixsans'
mpl.rcParams['mathtext.default'] = 'regular'

#--------------------------------------------------------------------
# Define transformation matricies
#--------------------------------------------------------------------
mirror_matrix = np.array([[-1, 0], [0, 1]])
flip_matrix = np.array([[1, 0], [0, -1]])


#--------------------------------------------------------------------
# Internal functions
#--------------------------------------------------------------------
def _angle(a, b):
    """ Compute angle from coordinate a to b
    """
    theta = np.degrees(np.arctan2(b[1] - a[1], (b[0] - a[0])))
    return theta


def _distance(a, b):
    ''' Compute distance from A to B '''
    r = np.sqrt((a[0]-b[0])**2 + (a[1]-b[1])**2)
    return r


def _merge_elements(elm_def):
    """ Combine element with it's base elements.
    """
    while 'base' in elm_def:
        base_elm = elm_def['base']
        elm_def.pop('base')

        elm_def['paths'] = elm_def.get('paths', []) + base_elm.get('paths',  [])
        elm_def['labels'] = elm_def.get('labels', []) + base_elm.get('labels', [])
        elm_def['shapes'] = elm_def.get('shapes', []) + base_elm.get('shapes', [])

        # New anchors with same name should overwrite old
        d = base_elm.get('anchors', {}).copy()
        d.update(elm_def.get('anchors', {}))
        elm_def['anchors'] = d

        for i in base_elm:
            # Everything that's not a list. New key/values overwrite old.
            # If base has a base, will be added and loop will pick it up.
            if i not in ['paths', 'labels', 'shapes', 'anchors'] and i not in elm_def:
                elm_def[i] = base_elm[i]
    return elm_def


#--------------------------------------------------------------------
def group_elements(drawing, anchors=None):
    ''' Combine all elements in a drawing into a single element that can be added to
        another drawing. Returns an element definition.

        drawing: The drawing object. All elements in the drawing will be combined.
        anchors: New anchor dictionary to use for the new element.
    '''
    new_elm = {'paths': [],
               'shapes': [],
               'labels': [],
               'anchors': {}}
    for elm in drawing._elm_list:
        new_elm['paths'].extend(elm.paths)
        for s in elm.shapes:
            s = s.copy()
            if s['shape'] == 'circle':
                s['center'] = elm.translate(s['center'] - elm.ofst)
            elif s['shape'] == 'poly':
                s['xy'] = elm.translate(s['xy'] - elm.ofst)
            elif s['shape'] == 'arc':
                s['center'] = elm.translate(s['center'] - elm.ofst)
            elif s['shape'] == 'arrow':
                s['start'] = elm.translate(s['start'] - elm.ofst)
                s['end'] = elm.translate(s['start'] - elm.ofst)
            new_elm['shapes'].append(s)

        if anchors is not None:
            new_elm['anchors'] = anchors

        for label, loc, align, size in elm.strs:
            loc = elm.translate(np.array(loc))
            new_elm['labels'].append({'label':label, 'pos':loc, 'align':align, 'size':size})
        new_elm['extend'] = False
        new_elm['drop'] = drawing.here

    return new_elm


#--------------------------------------------------------------------
# Main drawing class
#--------------------------------------------------------------------
class Drawing(object):
    def __init__(self, unit=3.0, inches_per_unit=0.5, txtofst=0.1, fontsize=16, font='sans-serif', color='black', lw=1.5):
        """ Set up a new circuit drawing.
    unit : Full length of a resistor element in matplotlib plot units.
           Inner portion of resistor is length 1.
    inches_per_unit : inches per unit to scale drawing.
    txtofst  : Default distance from element to text label
    fontsize : Default font size
    font     : matplotlib font-family
    color    : default color for elements
    lw       : default line width
    """
        # Default values
        self.unit = unit  # Default length of 2-terminal element, including leads
        self.inches_per_unit = inches_per_unit
        self.txtofst = txtofst
        self.fontsize = fontsize
        self.font = font
        self.color = color
        self.lw = lw

        # State variables
        self.here = np.array([0, 0])
        self.theta = 0
        self._state = []
        self._elm_list = []

    def add(self, elm_def, **kwargs):
        ''' Add an element to the schematic.

elm_def  : element definition dictionary (see below)

Element properties are specified using kwargs as follows.

Position of element [Default = endpoint of last element]:
    xy     : [x,y] starting coordiante.
             Element drawn in current direction and default length.
    endpts : [[x1,y1], [x2,y2]] start and end coordinates
    to     : [x,y] end coordinate
    tox    : x-value of end coordinate (y-value same as start)
    toy    : y-value of end coordinate (x-value same as start)
    l      : total length of element
    zoom   : zoom/magnification for element (default=1)
    anchor : 'xy' argument refers to this position within the element.
             For example, an opamp can be anchored to 'in1', 'in2', or 'out'

'to', 'tox', 'toy', 'l' can be used with 'xy' to define start and end points.

Direction [Default = direction of last element]:
    d       : direction ['up','down','left','right']
    theta   : angle (in degrees) to draw the element. Overrides 'd'.
    flip    : flip the element vertically (when theta=0)
    reverse : reverse a directional element (e.g. diode)

Labels [Default = no label]:
    label, toplabel, botlabel, lftlabel, rgtlabel:
        Add a string label to the element. Can be a string, or list
        of strings to be evenly spaced along element, e.g. ['-','R1','+']
        Use $ for latex-style symbols, e.g. '$R_1 = 100 \Omega$'

    lblofst : offset between text label and element

Other:
    move_cur : move the cursor after drawing. Default=True.
    color    : matplotlib color for element. e.g. 'red', '#34a4e6', (.8,0,.8)
    '''
        e = Element(elm_def, self, **kwargs)
        self._elm_list.append(e)
        if kwargs.get('move_cur', True) and elm_def.get('move_cur', True):
            self.here = e.here
            self.theta = e.theta
        return e

    def push(self):
        """ Push/save the drawing state. Location and angle are saved.
        """
        self._state.append((self.here, self.theta))

    def pop(self):
        """ Pop/load the drawing state.
            Location and angle are returned to previously pushed state.
        """
        if len(self._state) > 0:
            self.here, self.theta = self._state.pop()

    def draw(self, ax=None, showframe=False, showplot=True):
        """ Draw the diagram.
            ax : matplotlib axis to draw to.
            showframe : Show the plot frame/axis. Useful for debugging.
            showplot : Show the plot in matplotlib window in non-interactive mode.
        """

        mpl.rcParams['font.size'] = self.fontsize
        mpl.rcParams['font.family'] = self.font

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        for e in self._elm_list:
            e.draw(ax)

        ax.autoscale_view(True)  # This autoscales all the shapes too
        # NOTE: arrows don't seem to be included in autoscale!
        xlim = np.array(ax.get_xlim())
        ylim = np.array(ax.get_ylim())
        xlim[0] = xlim[0]-.1   # Add a .1 unit border to pick up lost pixels
        ylim[0] = ylim[0]-.1
        xlim[1] = xlim[1]+.1
        ylim[1] = ylim[1]+.1
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        w = xlim[1]-xlim[0]
        h = ylim[1]-ylim[0]

        if not showframe:
            ax.axes.get_xaxis().set_visible(False)
            ax.axes.get_yaxis().set_visible(False)
            ax.set_frame_on(False)

        self.ax = ax
        self.fig = ax.get_figure()
        if not plt.isinteractive() and showplot:
            plt.show()

        # Grow the figure size so that elements are always the same
        # Do after show() because it messes with size.
        ax.get_figure().set_size_inches(self.inches_per_unit*w, self.inches_per_unit*h)

    def save(self, fname, transparent=True, dpi=72):
        """ Save figure to file.
            fname : filename to save.
                    File type automatically determined from extension.
        """
        self.fig.savefig(fname, bbox_extra_artists=self.ax.get_default_bbox_extra_artists(), bbox_inches='tight', transparent=transparent, dpi=dpi)

    # Functions to help define specific elements
    def labelI(self, elm, label='', arrowofst=0.4, arrowlen=2, reverse=False, top=True):
        ''' Add an arrow element along side another element
        elm       : element to add arrow to
        label     : string or list of strings to space along arrow
        arrowofst : distance from element to arrow
        arrowlen  : length of arrow in drawing units
        reverse   : reverse the arrow direction
        top       : [True,False] draw on top (or bottom) of element
         '''
        if not top:
            arrowofst = -arrowofst

        x = elm.center[0] - np.sin(np.deg2rad(elm.theta))*arrowofst
        y = elm.center[1] + np.cos(np.deg2rad(elm.theta))*arrowofst

        if reverse:
            arrowlen = -arrowlen

        arrow = copy.deepcopy(elements.ARROW_I)
        arrow['shapes'][0]['start'] = [1-arrowlen/2, 0]
        arrow['shapes'][0]['end'] = [1+arrowlen/2, 0]

        if top:
            self.add(arrow, theta=elm.theta, label=label, anchor='center', xy=[x, y])
        else:
            self.add(arrow, theta=elm.theta, botlabel=label, anchor='center', xy=[x, y])

    def loopI(self, elm_list, label='', d='cw', theta1=35, theta2=-35, pad=.2):
        ''' Label the loop current bordered by elements in list,
            elm_list : boundary elements in order of top, right, bot, left
            label    : text label for center of loop
            d        : arrow direction 'cw' or 'ccw'
            theta1   : start angle of arrow arc (degrees)
            theta2   : end angle of arrow arc (degrees)
            pad      : distance between element and arc
        '''
        bbox1 = elm_list[0].translate([elm_list[0].xmin, elm_list[0].ymin])
        bbox2 = elm_list[0].translate([elm_list[0].xmax, elm_list[0].ymax])
        top = min(bbox1[1], bbox2[1])
        bbox1 = elm_list[1].translate([elm_list[1].xmin, elm_list[1].ymin])
        bbox2 = elm_list[1].translate([elm_list[1].xmax, elm_list[1].ymax])
        rght = min(bbox1[0], bbox2[0])
        bbox1 = elm_list[2].translate([elm_list[2].xmin, elm_list[2].ymin])
        bbox2 = elm_list[2].translate([elm_list[2].xmax, elm_list[2].ymax])
        bot = max(bbox1[1], bbox2[1])
        bbox1 = elm_list[3].translate([elm_list[3].xmin, elm_list[3].ymin])
        bbox2 = elm_list[3].translate([elm_list[3].xmax, elm_list[3].ymax])
        left = max(bbox1[0], bbox2[0])

        top = top - pad
        bot = bot + pad
        rght = rght - pad
        left = left + pad
        center = [(left+rght)/2, (top+bot)/2]

        loop = {'shapes': [{'shape': 'arc',
                            'center': [0,0],
                            'theta1': theta1,
                            'theta2': theta2,
                            'width': rght-left,
                            'height': top-bot,
                            'arrow': d}],
                'move_cur': False,
                'lblloc': 'center',
                'lblofst': 0
                }

        L = self.add(loop, xy=center, d='right')
        L.add_label(label, loc='center', ofst=0, align=('center', 'center'))
        return L


#--------------------------------------------------------------------
# Circuit element class
#--------------------------------------------------------------------
class Element(object):
    def __init__(self, elm_def, drawing, **kwargs):
        """ Initialize the element.
elm_def  : element definition dictionary (see below)
drawing  : drawing to add the element to

Element properties are specified using kwargs as follows.

Position of element [Default = endpoint of last element]:
    xy     : [x,y] starting coordiante.
             Element drawn in current direction and default length.
    endpts : [[x1,y1], [x2,y2]] start and end coordinates
    to     : [x,y] end coordinate
    tox    : x-value of end coordinate (y-value same as start)
    toy    : y-value of end coordinate (x-value same as start)
    l      : total length of element
    zoom   : zoom/magnification for element (default=1)
    anchor : 'xy' argument refers to this position within the element.
             For example, an opamp can be anchored to 'in1', 'in2', or 'out'

'to', 'tox', 'toy', 'l' can be used with 'xy' to define start and end points.

Direction [Default = direction of last element]:
    d       : direction ['up','down','left','right']
    theta   : angle (in degrees) to draw the element. Overrides 'd'.
    flip    : flip the element vertically (when theta=0)
    reverse : reverse a directional element (e.g. diode)

Labels [Default = no label]:
    label, toplabel, botlabel, lftlabel, rgtlabel:
        Add a string label to the element. Can be a string, or list
        of strings to be evenly spaced along element, e.g. ['-','R1','+']
        Use $ for latex-style symbols, e.g. '$R_1 = 100 \Omega$'

    lblofst : offset between text label and element

Other:
    move_cur : move the cursor after drawing. Default=True.
    color    : matplotlib color for element. e.g. 'red', '#34a4e6', (.8,0,.8)
    ls       : line style. same as matplotlib: ['-', ':', '--']
    lw       : line width. same as matplotlib. Default=1
    """
        # Flatten element def with base elements
        self.defn = elm_def.copy()
        if 'base' in self.defn:
            self.defn = _merge_elements(self.defn)

        # Get some parameters
        self.drawing = drawing
        self.shapes = self.defn.get('shapes', [])
        self.color = kwargs.get('color', self.defn.get('color', drawing.color))
        self.ls = kwargs.get('ls', self.defn.get('ls', '-'))
        self.lw = kwargs.get('lw', self.defn.get('lw', drawing.lw))
        totlen = kwargs.get('l', drawing.unit)

        # Determine theta angle of element
        default_theta = self.defn.get('theta', drawing.theta)
        if 'endpts' in kwargs:
            endpts = kwargs['endpts']
            self.theta = _angle(endpts[0], endpts[1])
        elif 'theta' in kwargs:
            self.theta = kwargs['theta']
        elif 'd' in kwargs:
            self.theta = {'up':90, 'right':0, 'left':180, 'down':270}[kwargs['d']]
        elif 'to' in kwargs:
            self.theta = _angle(kwargs.get('xy', drawing.here), kwargs['to'])
        else:
            self.theta = default_theta

        # Get offset to element position
        if 'endpts' in kwargs:
            self.trans = kwargs['endpts'][0]
        elif 'xy' in kwargs:
            self.trans = np.array(kwargs['xy'])
        else:
            self.trans = drawing.here

        # Get offset due to anchoring
        if 'anchor' in kwargs:
            anchor = kwargs['anchor']
            if 'anchors' in self.defn and anchor in self.defn['anchors']:
                self.ofst = np.array(self.defn['anchors'][anchor])
            else:
                print('Anchor %s not defined in element' % anchor)
                self.ofst = np.array([0, 0])  # Default to 'start'
        else:
            self.ofst = np.array([0, 0])

        # Last path in list should be the one to get lead extensions.
        # The others should just have extend offset applied.
        self.path_def = [np.array(p) for p in self.defn.get('paths', [])]
        self.paths = []   # Translated paths

        if len(self.path_def) == 0:
            start = np.array([0, 0])
            end = np.array([0, 0])
        elif self.defn.get('extend', True):
            in_path = self.path_def[-1]
            dz = in_path[-1]-in_path[0]
            in_len = np.sqrt(dz[0]*dz[0]+dz[1]*dz[1])

            if 'endpts' in kwargs:
                endpts = kwargs['endpts']
                totlen = _distance(endpts[0], endpts[1])
            elif 'to' in kwargs:
                # Move until X or Y position is 'end'. Depends on direction
                endpt = kwargs['to']
                totlen = _distance(self.trans, endpt)
            elif 'tox' in kwargs:
                # Allow either full coordinate (only keeping x), or just an x value
                if isinstance(kwargs['tox'], float) or isinstance(kwargs['tox'], int):
                    x = float(kwargs['tox'])
                else:
                    x = kwargs['tox'][0]
                endpt = [x, self.trans[1]]
                totlen = _distance(self.trans, endpt)
            elif 'toy' in kwargs:
                # Allow either full coordinate (only keeping y), or just a y value
                if isinstance(kwargs['toy'], float) or isinstance(kwargs['toy'], int):
                    y = kwargs['toy']
                else:
                    y = kwargs['toy'][1]
                endpt = [self.trans[0], y]
                totlen = _distance(self.trans, endpt)

            lead_len = (totlen - in_len)/2
            start = in_path[0] - np.array([lead_len, 0])
            end = in_path[-1] + np.array([lead_len, 0])
            p = np.vstack((start, in_path, end))
            if _distance(self.ofst, [0, 0]) == 0:
                self.ofst = start  # Offset due to lead extension
            p = p - self.ofst
            self.paths.append(p)
        else:
            # Don't extend leads
            self.paths.append(self.path_def[-1] - self.ofst)

        # Process all the other paths, adding in extend offset
        for p in self.path_def[:-1]:
            self.paths.append(p - self.ofst)

        if len(self.path_def) == 0:
            self.drop = np.array([0, 0])
        else:
            self.drop = self.defn.get('drop', self.paths[0][-1])

        self.flip = kwargs.get('flip', False)       # Flip (up/down when theta=0)
        self.reverse = kwargs.get('reverse', False) # Reverse (left/right)

        self.xmin, self.ymin, self.xmax, self.ymax = self.get_bbox()

        # Get labels from arguments.
        # 'label' defaults to 'top' unless overridden in element def.
        self.strs = []
        if 'base' not in self.defn:
            label = kwargs.get('label', None)
            dfltlbl = kwargs.get('lblloc', self.defn.get('lblloc', 'top'))
            toplabel = kwargs.get('toplabel', None)
            botlabel = kwargs.get('botlabel', None)
            rgtlabel = kwargs.get('rgtlabel', None)
            lftlabel = kwargs.get('lftlabel', None)
            self.lbl_ofst = kwargs.get('lblofst', self.defn.get('lblofst', drawing.txtofst))
            if label is not None:    self.add_label(label, dfltlbl,  ofst=self.lbl_ofst, size=drawing.fontsize)
            if toplabel is not None: self.add_label(toplabel, 'top', ofst=self.lbl_ofst, size=drawing.fontsize)
            if botlabel is not None: self.add_label(botlabel, 'bot', ofst=self.lbl_ofst, size=drawing.fontsize)
            if rgtlabel is not None: self.add_label(rgtlabel, 'rgt', ofst=self.lbl_ofst, size=drawing.fontsize)
            if lftlabel is not None: self.add_label(lftlabel, 'lft', ofst=self.lbl_ofst, size=drawing.fontsize)

        txtlist = self.defn.get('labels', [])
        for txtlbl in txtlist:
            align = txtlbl.get('align', ('center', 'center'))
            size = txtlbl.get('size', drawing.fontsize)
            pos = txtlbl.get('pos', [0, 0])
            self.strs.append(((txtlbl['label'], np.array(pos)-self.ofst, align, size)))

        self.z = kwargs.get('zoom', 1)

        # Generate rotation matrix from theta
        c = np.cos(np.radians(self.theta))
        s = np.sin(np.radians(self.theta))
        self.m = np.array([[c, s],[-s, c]])

        # Translate paths
        for i in range(len(self.paths)):
            self.paths[i] = self.translate(self.paths[i])

        # Define anchors
        if len(self.path_def) == 0:
            setattr(self, 'start',  self.translate(np.array([0, 0])))
            setattr(self, 'end',    self.translate(np.array([0, 0])))
            setattr(self, 'center', self.translate(np.array([0, 0])))
        else:
            setattr(self, 'start', self.paths[0][0])
            setattr(self, 'end',   self.paths[0][-1])
            setattr(self, 'center', (self.paths[0][0] + self.paths[0][-1])/2)
        if 'anchors' in self.defn:
            for aname, apoint in self.defn['anchors'].items():
                if getattr(self, aname, None) is not None:  # Try not to clobber element parameter names!
                    aname = 'anchor_' + aname
                setattr(self, aname, self.translate(np.array(apoint)-self.ofst))

        # Translated end point for new drawing position
        self.here = self.translate(self.drop, doflip=False)

    def get_bbox(self):
        """ Get element bounding box, including path and shapes.
        """
        xmin = ymin = np.inf
        xmax = ymax = -np.inf
        for path in self.paths:
            for p in path:
                xmin = min(xmin, p[0])
                ymin = min(ymin, p[1])
                xmax = max(xmax, p[0])
                ymax = max(ymax, p[1])
        for s in self.shapes:
            if s['shape'] == 'arrow':
                start = np.array(s['start']) - self.ofst
                end   = np.array(s['end']) - self.ofst
                xmin = min(xmin, start[0], end[0])
                ymin = min(ymin, start[1], end[1])
                xmax = max(xmax, start[0], end[0])
                ymax = max(ymax, start[1], end[1])
            elif s['shape'] == 'circle':
                center = np.array(s['center']) - self.ofst
                rad    = s['radius']
                xmin = min(xmin, center[0]-rad)
                ymin = min(ymin, center[1]-rad)
                xmax = max(xmax, center[0]+rad)
                ymax = max(ymax, center[1]+rad)
            elif s['shape'] == 'poly':
                for v in s['xy']:
                    v = v - self.ofst
                    xmin = min(xmin, v[0])
                    ymin = min(ymin, v[1])
                    xmax = max(xmax, v[0])
                    ymax = max(ymax, v[1])
            elif s['shape'] == 'arc':
                center = s['center'] - self.ofst
                xmin = min(xmin, center[0] - s['width'])
                ymin = min(ymin, center[1] - s['height'])
                xmax = max(xmax, center[0] + s['width'])
                ymax = max(ymax, center[1] + s['height'])

        # Don't want to propogate infinities (e.g. shape not defined above)
        if xmax == -np.Inf: xmax = 0
        if ymax == -np.Inf: ymax = 0
        if xmin ==  np.Inf: xmin = 0
        if ymin ==  np.Inf: ymin = 0
        return xmin, ymin, xmax, ymax

    def add_label(self, label, loc='top', ofst=None, align=None, size=None):
        """ Add a label at the appropriate position
            label: text label to add
            loc: location for the label ['top', 'bot', 'lft', 'rgt'].
            txtofst: unit offset between element bounding box and label.
            align: alignment tuple for (horizontal, vertical):
                   (['center', 'left', 'right'], ['center', 'top', 'bottom'])
            size: font size
        """
        if isinstance(ofst, list) and loc is not 'center':
            raise TypeError('Offset must not be list for loc=%s'%loc)

        if ofst is None:
            ofst = self.lbl_ofst

        # This ensures a 'top' label is always on top, regardless of rotation
        if (self.theta % 360) > 90 and (self.theta % 360) <= 270:
            if loc == 'top':
                loc = 'bot'
            elif loc == 'bot':
                loc = 'top'
            elif loc == 'lft':
                loc = 'rgt'
            elif loc == 'rgt':
                loc = 'lft'

        if align is None:   # Determine best alignment for label based on angle
            th = self.theta
            # Below alignment divisions work for label on top. Rotate angle for other sides.

            if loc == 'lft':
                th = th + 90
            elif loc == 'bot':
                th = th + 180
            elif loc == 'rgt':
                th = th + 270
            th = th % 360  # Normalize angle so it's positive, clockwise

            if th < 22.5:         # 0 to +22 deg
                align = ('center', 'bottom')
            elif th < 22.5+45:    #22 to 67
                align = ('right', 'bottom')
            elif th < 22.5+45*2:  # 67 to 112
                align = ('right', 'center')
            elif th < 22.5+45*3:  # 112 to 157
                align = ('right', 'top')
            elif th < 22.5+45*4:  # 157 to 202
                align = ('center', 'top')
            elif th < 22.5+45*5:  # 202 to 247
                align = ('left', 'top')
            elif th < 22.5+45*6:  # 247 to 292
                align = ('left', 'center')
            elif th < 22.5+45*7:  # 292 to 337
                align = ('left', 'bottom')
            else:                 #337 to 0
                align = ('center', 'bottom')

        xmax = self.xmax
        xmin = self.xmin
        ymax = self.ymax
        ymin = self.ymin

        if size is None:
            size=self.drawing.fontsize

        if self.flip:
            ymax, ymin = ymin, ymax
            if loc == 'top' or loc == 'bot':
                ofst = -ofst
        if self.reverse:
            xmax, xmin = xmin, xmax
            if loc == 'lft' or loc == 'rgt':
                ofst = -ofst

        if isinstance(label, list):
            # Divide list along length
            if loc == 'top':
                for i, lbltxt in enumerate(label):
                    xdiv = (xmax-xmin)/(len(label)+1)
                    lbl_pt = [xmin+xdiv*(i+1), ymax+ofst]
                    self.strs.append((label[i], lbl_pt, align, size))
            elif loc == 'bot':
                for i, lbltxt in enumerate(label):
                    xdiv = (xmax-xmin)/(len(label)+1)
                    lbl_pt = [xmin+xdiv*(i+1), ymin-ofst]
                    self.strs.append((label[i], lbl_pt, align, size))
            elif loc == 'lft':
                for i, lbltxt in enumerate(label):
                    ydiv = (ymax-ymin)/(len(label)+1)
                    lbl_pt = [xmin-ofst, ymin+ydiv*(i+1)]
                    self.strs.append((label[i], lbl_pt, align, size))
            elif loc == 'rgt':
                for i, lbltxt in enumerate(label):
                    ydiv = (ymax-ymin)/(len(label)+1)
                    lbl_pt = [xmax+ofst, ymin+ydiv*(i+1)]
                    self.strs.append((label[i], lbl_pt, align, size))
            elif loc == 'center':
                for i, lbltxt in enumerate(label):
                    xdiv = (xmax-xmin)/(len(label)+1)
                    lbl_pt = [xmin+xdiv*(i+1), ofst]
                    self.strs.append((label[i], lbl_pt, align, size))
        elif isinstance(label, str):
            # Place in center
            if loc == 'top':
                lbl_pt = [(xmax+xmin)/2, ymax+ofst]
            elif loc == 'bot':
                lbl_pt = [(xmax+xmin)/2, ymin-ofst]
            elif loc == 'lft':
                lbl_pt = [xmin-ofst, (ymax+ymin)/2]
            elif loc == 'rgt':
                lbl_pt = [xmax+ofst, (ymax+ymin)/2]
            elif loc == 'center':
                if isinstance(ofst, list):
                    lbl_pt = [(xmax+xmin)/2+ofst[0], (ymax+ymin)/2+ofst[1]]
                else:
                    lbl_pt = [(xmax+xmin)/2, (ymax+ymin)/2+ofst]
            self.strs.append((label, lbl_pt, align, size))

    def translate(self, pt, doflip=True):
        """ Translate the coordinate(s) to drawing position based on
            element's parameters.
            pt     : single coordinate or array of coordinates
            doflip : translate includes mirror/flip?
        """
        pt = np.array(pt)
        if doflip and self.reverse:
            pt = np.dot(pt, mirror_matrix) + [self.drop[0], 0]
        if doflip and self.flip:
            pt = np.dot(pt, flip_matrix)
        return np.dot(pt*self.z, self.m) + self.trans

    def draw(self, ax, showframe=False):
        """ Draw the element
            ax        : matplotlib axis
            showframe : Draw the axis frame. Useful for debugging.
        """
        for path in self.paths:
            ax.plot(path[:, 0], path[:, 1], color=self.color, lw=self.lw,
                    solid_capstyle='round', ls=self.ls)

        for s in self.shapes:
            if s.get('shape') == 'circle':
                xy = np.array(s.get('center', [0, 0]))
                xy = self.translate(xy - self.ofst)
                rad = s.get('radius', 1) * self.z
                fill = s.get('fill', False)
                fillcolor = s.get('fillcolor', self.color)
                circ = plt.Circle(xy=xy, radius=rad, ec=self.color,
                                  fc=fillcolor, fill=fill, zorder=3, lw=self.lw)
                ax.add_patch(circ)
            elif s.get('shape') == 'poly':
                xy = np.array(s.get('xy', [[0, 0]]))
                xy = self.translate(xy - self.ofst)
                closed = s.get('closed', True)
                fill = s.get('fill', False)
                fillcolor = s.get('fillcolor', self.color)
                poly = plt.Polygon(xy=xy, closed=closed, ec=self.color,
                                   fc=fillcolor, fill=fill, zorder=3, lw=self.lw)
                ax.add_patch(poly)
            elif s.get('shape') == 'arc':
                xy = np.array(s.get('center', [0, 0]))
                xy = self.translate(xy - self.ofst)

                w = s.get('width', 1) * self.z
                h = s.get('height', 1) * self.z
                th1 = s.get('theta1', 35)
                th2 = s.get('theta2', -35)
                if self.reverse: th1, th2 = th2, th1

                angle = s.get('angle', self.theta)
                arc = Arc(xy, width=w, height=h, theta1=th1,
                          theta2=th2, angle=angle, color=self.color, lw=self.lw)
                ax.add_patch(arc)

                # Add an arrowhead to the arc
                arrow = s.get('arrow', None)  # 'cw' or 'ccw' or None
                if arrow is not None:
                    if arrow == 'ccw':
                        dx = np.cos(np.deg2rad(th2+90)) / 100
                        dy = np.sin(np.deg2rad(th2+90)) / 100
                        s = [xy[0] + w/2*np.cos(np.deg2rad(th2)),
                             xy[1] + h/2*np.sin(np.deg2rad(th2))]
                    else:
                        dx = -np.cos(np.deg2rad(th1+90)) / 100
                        dy = - np.sin(np.deg2rad(th1+90)) / 100

                        s = [xy[0] + w/2*np.cos(np.deg2rad(th1)),
                             xy[1] + h/2*np.sin(np.deg2rad(th1))]

                    # Rotate the arrow head
                    co = np.cos(np.radians(angle))
                    so = np.sin(np.radians(angle))
                    m = np.array([[co, so],[-so, co]])
                    s = np.dot(s-xy, m)+xy
                    darrow = np.dot([dx, dy], m)

                    ax.arrow(s[0], s[1], darrow[0], darrow[1], head_width=.15,
                             head_length=.25, color=self.color)

            elif s.get('shape') == 'arrow':
                start = np.array(s.get('start', [0, 0]))
                end = np.array(s.get('end', [1, 0]))
                start = self.translate(start - self.ofst)
                end = self.translate(end - self.ofst)
                hw = s.get('headwidth', .1)
                hl = s.get('headlength', .2)
                ax.arrow(start[0], start[1], end[0]-start[0], end[1]-start[1],
                         head_width=hw, head_length=hl,
                         length_includes_head=True, color=self.color, lw=self.lw)

        for label, loc, align, size in self.strs:
            loc = self.translate(np.array(loc))
            ha = align[0]
            va = align[1]
            ax.text(loc[0], loc[1], label, transform=ax.transData,
                    horizontalalignment=ha, verticalalignment=va,
                    fontsize=size)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name = 'SchemDraw',
    packages = ['SchemDraw'],
    version = '0.3.1a',
    description = 'Electrical circuit schematic drawing',
    author = 'Collin J. Delker',
    author_email = 'developer@collindelker.com',
    url = 'http://cdelker.bitbucket.io/SchemDraw',
    keywords = ['circuit', 'schematic', 'electrical'],
    install_requires=['numpy', 'matplotlib'],
    classifiers = [
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: End Users/Desktop',
    ],
)
